@@ -1,4 +1,4 @@
-# argumentor - A library to work with command-line arguments
+# argumentor - A simple, copylefted, lightweight library to work with command-line arguments in Python
 # Copyright (C) 2021 Twann <twann@ctemplar.com>
 
 # This program is free software: you can redistribute it and/or modify
@@ -20,21 +20,27 @@ import re
 from .exceptions import ExecNameError, OperationExistsError, OptionExistsError, ArgumentValueError, GroupExistsError, \
     GroupNotExistsError, InvalidOptionError
 
-__version__ = '0.0.2'
+__version__ = '0.0.3'
 
 
 class Arguments:
 
-    def __init__(self, arguments: list = None, exec_name: str = None, generate_help_page: bool = True,
-                 one_operation_required: bool = True, allow_short_args_together: bool = True) -> None:
+    def __init__(self, arguments: list = None, exec_name: str = None, help_page: str = None,
+                 one_operation_required: bool = True, allow_short_args_together: bool = True,
+                 show_operation_groups_on_help_page: bool = True, show_option_groups_on_help_page: bool = True,
+                 copyright_header: str = None, more_info_footer: str = None) -> None:
         """
         Argumentor parser object
         :param arguments: Optional. Arguments to parse. If not set, sys.argv[1:] will be used
         :param exec_name: Optional. The name of the executable running. If not set, sys.argv[0] will be used
-        :param generate_help_page: Defaults to True. Generate a help page as self.help_page
+        :param help_page: Optional. Specify a custom help page as self.help_page. If None, argumentor will generate one
         :param one_operation_required: Defaults to True. If no operation is specified, the script will return an error
         :param allow_short_args_together: Defaults to True. Allow short arguments to be combined together
+        :param show_operation_groups_on_help_page: Defaults to True. Group operations by group on generated help page
+        :param show_option_groups_on_help_page: Defaults to True. Group options by group on generated help page
         (ex. '-vz' instead of '-v -z')
+        :param copyright_header: Optional. Specify a copyright header to be shown on generated help page
+        :param more_info_footer: Optional. Add a custom message at the end of the generated help page
         """
 
         if arguments is None:
@@ -49,13 +55,12 @@ class Arguments:
         else:
             self.exec_name = exec_name
 
-        self.generate_help_page = generate_help_page                    # Generate a help page
         self.one_operation_required = one_operation_required            # One operation is required
         self.allow_short_args_together = allow_short_args_together      # Allow short args to be combined
 
         self.operations = {}                                            # Available operations
         self.short_operations = {}                                      # Available short operations
-        self.operation_groups = []                                      # Existing operation groups
+        self.operation_groups = {}                                      # Existing operation groups
         self.options = {}                                               # Available options
         self.short_options = {}                                         # Available short options
         self.option_groups = {}                                         # Existing option groups
@@ -66,6 +71,105 @@ class Arguments:
         self.output_operations = {}                                     # Operations result after parsing
         self.output_options = {}                                        # Options result after parsing
 
+        self.show_operation_groups_on_help_page = show_operation_groups_on_help_page
+        self.show_option_groups_on_help_page = show_option_groups_on_help_page
+        self.copyright_header = copyright_header
+        self.more_info_footer = more_info_footer
+
+        if help_page:
+            self.help_page = help_page
+            self.generate_help_page = False
+        else:
+            self.help_page = self.__generate_help_page()
+            self.generate_help_page = True
+
+    def __generate_help_page(self) -> str:
+        """
+        :return: str: The help page
+        """
+        output = ''
+        if self.copyright_header is not None:
+            output += f'{self.copyright_header}\n'
+
+        if self.operations is not None and self.options is None:
+            output += f'usage: {self.exec_name} [operation] [arg(s)]\n\n'
+        elif self.operations is None and self.options is not None:
+            output += f'usage: {self.exec_name} [option(s)] [arg(s)]\n\n'
+        elif self.operations is not None and self.options is not None:
+            output += f'usage: {self.exec_name} [operation] [option(s)] [arg(s)]\n\n'
+        else:
+            output += f'usage: {self.exec_name} [arg(s)]\n\n'
+
+        output += '<operations>\n'
+
+        for group in self.operation_groups.keys():
+            if self.show_operation_groups_on_help_page:
+                output += f'* {group}\n'
+            for operation in self.__get_operations_from_group(group):
+                line = ''
+                if self.operations[operation]['short_operation'] is not None:
+                    line += f'  {self.operations[operation]["short_operation"]}'
+                line += f'  {operation}'
+                if len(line) > 15:
+                    line += f'\t\t\t{self.operations[operation]["description"]}\n'
+                else:
+                    line += f'\t\t\t\t{self.operations[operation]["description"]}\n'
+                output += line
+        for operation in self.operations:
+            if self.operations[operation]['group'] is None:
+                line = ''
+                if self.operations[operation]['short_operation'] is not None:
+                    line += f'  {self.operations[operation]["short_operation"]}'
+                line += f'  {operation}'
+                if len(line) > 15:
+                    line += f'\t\t\t{self.operations[operation]["description"]}\n'
+                else:
+                    line += f'\t\t\t\t{self.operations[operation]["description"]}\n'
+                output += line
+
+        output += '\n<options>\n'
+
+        for group in self.option_groups.keys():
+            if self.show_option_groups_on_help_page:
+                output += f'* {group}\n'
+            for option in self.__get_options_from_group(group):
+                line = ''
+                if self.options[option]['short_option'] is not None:
+                    line += f'  {self.options[option]["short_option"]}'
+                line += f'  {option}'
+                if len(line) > 15:
+                    line += f'\t\t\t{self.options[option]["description"]}\n'
+                else:
+                    line += f'\t\t\t\t{self.options[option]["description"]}\n'
+                output += line
+        for option in self.options:
+            if self.options[option]['group'] is None:
+                line = ''
+                if self.options[option]['short_operation'] is not None:
+                    line += f'  {self.options[option]["short_option"]}'
+                line += f'  {option}'
+                if len(line) > 15:
+                    line += f'\t\t\t{self.options[option]["description"]}\n'
+                else:
+                    line += f'\t\t\t\t{self.options[option]["description"]}\n'
+                output += line
+
+        if self.more_info_footer is not None:
+            output += f'\n{self.more_info_footer}'
+        return output
+
+    def __get_operations_from_group(self, group: str) -> list:
+        if group not in self.operation_groups:
+            raise GroupNotExistsError(f'operation group "{group}" does not exist')
+        else:
+            return self.operation_groups[group]['operations']
+
+    def __get_options_from_group(self, group: str) -> list:
+        if group not in self.option_groups:
+            raise GroupNotExistsError(f'option group "{group}" does not exist')
+        else:
+            return self.option_groups[group]['options']
+
     def add_operation(self, operation: str, value_type, description: str = None,
                       short_operation: str = None, required: bool = False, group: str = None) -> None:
         """
@@ -84,26 +188,31 @@ class Arguments:
         elif operation in self.options.keys():
             raise OptionExistsError(f'option "{operation}" already exists')
         elif short_operation is not None and short_operation in self.short_operations.keys():
-            raise OperationExistsError(f'short operation "{operation}" already exists')
+            raise OperationExistsError(f'short operation "{short_operation}" already exists')
         elif short_operation is not None and short_operation in self.short_options.keys():
-            raise OptionExistsError(f'short option "{operation}" already exists')
+            raise OptionExistsError(f'short option "{short_operation}" already exists')
         elif value_type not in [bool, str, int, list]:
             raise ArgumentValueError(f'value type "{value_type}" is not supported')
         elif group is not None and group not in self.operation_groups:
             raise GroupNotExistsError(f'operation group "{group}" does not exist')
         elif short_operation is not None and len(short_operation) > 1 and not re.match(r'-[A-z0-9]', short_operation):
-            raise InvalidOptionError(f'short operation can only be one-character long when no dash, not "{operation}"')
+            raise InvalidOptionError(f'short operation can only be one-character long when no dash, not "{short_operation}"')
         elif short_operation is not None and len(short_operation) > 2:
-            raise InvalidOptionError(f'short operation can contain a dash and a character maximum, not "{operation}"')
+            raise InvalidOptionError(f'short operation can contain a dash and a character maximum, not "{short_operation}"')
         else:
             self.operations[operation] = {
                 'value_type': value_type,
                 'description': description,
                 'required': required,
-                'group': group
+                'group': group,
+                'short_operation': short_operation
             }
+            if group is not None:
+                self.operation_groups[group]['operations'].append(operation)
             if short_operation is not None:
                 self.short_operations[short_operation] = operation
+            if self.generate_help_page:
+                self.help_page = self.__generate_help_page()
 
     def add_operation_group(self, group_name: str) -> None:
         """
@@ -115,12 +224,12 @@ class Arguments:
         if group_name in self.operation_groups:
             raise GroupExistsError(f'operation group "{group_name}" already exists')
         else:
-            self.operation_groups.append(group_name)
+            self.operation_groups[group_name] = {'operations': []}
 
-    def add_options(self, option: str, value_type, description: str = None,
-                    short_option: str = None, required: bool = False, available_with_operations: list = None,
-                    required_by_operations: list = None, available_with_groups: list = None,
-                    required_by_groups: list = None, group: str = None, inherit_from_group: bool = True) -> None:
+    def add_option(self, option: str, value_type, description: str = None,
+                   short_option: str = None, required: bool = False, available_with_operations: list = None,
+                   required_by_operations: list = None, available_with_groups: list = None,
+                   required_by_groups: list = None, group: str = None, inherit_from_group: bool = True) -> None:
         """
         Add an option to the list of supported arguments
         :param option: The option (ex. '--force')
@@ -143,17 +252,17 @@ class Arguments:
         elif short_option is not None and short_option in self.short_options.keys():
             raise OptionExistsError(f'short option "{option}" already exists')
         elif short_option is not None and short_option in self.short_operations.keys():
-            raise OperationExistsError(f'short operation "{option}" already exists')
+            raise OperationExistsError(f'short operation "{short_option}" already exists')
         elif value_type not in [bool, str, int, list]:
             raise ArgumentValueError(f'value type "{value_type}" is not supported')
         elif group is not None and group not in self.option_groups.keys():
             raise GroupNotExistsError(f'option group "{group}" does not exist')
         elif not re.match(r'--[A-z0-9\-]*', option):
-            raise InvalidOptionError(f'option should start with two dashes: "--", not "{option}"')
+            raise InvalidOptionError(f'option should start with two dashes: "--", not "{short_option}"')
         elif short_option is not None and len(short_option) != 2:
-            raise InvalidOptionError(f'short option must be two-characters long, not "{option}"')
+            raise InvalidOptionError(f'short option must be two-characters long, not "{short_option}"')
         elif short_option is not None and not re.match(r'-[A-z0-9]', short_option):
-            raise InvalidOptionError(f'short option must start with a dash and finish with a character, not "{option}"')
+            raise InvalidOptionError(f'short option must start with a dash and finish with a character, not "{short_option}"')
         else:
             if group is not None and inherit_from_group:
                 self.options[option] = {
@@ -164,7 +273,8 @@ class Arguments:
                     'available_with_operations': self.option_groups[group]['available_with_operations'],
                     'required_by_operations': self.option_groups[group]['required_by_operations'],
                     'available_with_groups': self.option_groups[group]['available_with_groups'],
-                    'required_by_groups': self.option_groups[group]['required_by_groups']
+                    'required_by_groups': self.option_groups[group]['required_by_groups'],
+                    'short_option': short_option
                 }
                 if available_with_operations is not None:
                     self.options[option]['available_with_operations'] = available_with_operations
@@ -183,11 +293,15 @@ class Arguments:
                     'available_with_operations': available_with_operations,
                     'required_by_operations': required_by_operations,
                     'available_with_groups': available_with_groups,
-                    'required_by_groups': required_by_groups
+                    'required_by_groups': required_by_groups,
+                    'short_option': short_option
                 }
-
+            if group is not None:
+                self.option_groups[group]['options'].append(option)
             if short_option is not None:
                 self.short_options[short_option] = option
+            if self.generate_help_page:
+                self.help_page = self.__generate_help_page()
 
     def add_option_group(self, group_name: str, available_with_operations: list = None,
                          required_by_operations: list = None, available_with_groups: list = None,
@@ -209,7 +323,8 @@ class Arguments:
                 'available_with_operations': available_with_operations,
                 'required_by_operations': required_by_operations,
                 'available_with_groups': available_with_groups,
-                'required_by_groups': required_by_groups
+                'required_by_groups': required_by_groups,
+                'options': []
             }
 
     def __match_operation(self, operation_title: str, real_operation: str, real_argument: str) -> None:
@@ -247,15 +362,17 @@ class Arguments:
 
     def __match_option(self, option_title: str, real_option: str, real_argument: str) -> None:
 
-        if self.active_operation is not None and self.options[option_title]['available_with_operations'] is not None and not self.active_operation in self.options[option_title]['available_with_operations']:
+        if self.active_operation is not None and self.options[option_title]['available_with_operations'] is not None \
+                and self.active_operation not in self.options[option_title]['available_with_operations']:
             print(f'Error: option "{real_option}" is not available with operation "{self.active_operation}"')
             sys.exit(1)
 
-        elif self.active_operation is not None and self.options[option_title]['available_with_groups'] is not None and self.operations[self.active_operation]['group'] not in self.options[option_title]['available_with_groups']:
+        elif self.active_operation is not None and self.options[option_title]['available_with_groups'] is not None \
+                and self.operations[self.active_operation]['group'] not in self.options[option_title]['available_with_groups']:
             print(f'Error: option "{real_option}" is not available with operation "{self.active_operation}"')
             sys.exit(1)
 
-        elif self.active_operation is not None or not self.one_operation_required:
+        else:
 
             if self.options[option_title]['value_type'] == bool:
                 self.output_options[option_title] = True
@@ -325,7 +442,8 @@ class Arguments:
                     real_argument=argument
                 )
 
-            elif self.allow_short_args_together and len(argument) > 1 and (argument[0:1] in self.short_operations.keys()) or (argument[0:2] in self.short_operations.keys()):
+            elif self.allow_short_args_together and len(argument) > 1 and \
+                    (argument[0:1] in self.short_operations.keys() or argument[0:2] in self.short_operations.keys()):
 
                 for character in argument.replace('-', ''):
 
@@ -374,21 +492,27 @@ class Arguments:
 
             elif self.isolate_unmatched and self.active_operation is not None and not re.match(r'-[A-z0-9_.\-]*', argument):
 
-                self.output_operations[self.active_operation].append(argument)
+                if not self.operations[self.active_operation]['value_type'] == bool:
+                    self.output_operations[self.active_operation].append(argument)
 
             else:
 
                 print(f'Error: invalid argument "{argument}"')
                 sys.exit(1)
 
+        # Check if all active operations that require a list have at least one value
         for active_operation in self.output_operations.keys():
-            if self.operations[active_operation]['value_type'] == list and  self.output_operations[active_operation] == []:
+            if self.operations[active_operation]['value_type'] == list and self.output_operations[active_operation] == []:
                 print(f'Error: option "{active_operation}" requires at least one argument')
                 sys.exit(1)
+
+        # Check if all required operations are active
         for operation in self.operations:
             if self.operations[operation]['required'] and operation not in self.output_operations.keys():
                 print(f'Error: operation "{operation}" is required')
                 sys.exit(1)
+
+        # Check if all required options (by an operation or by the program) are available
         for option in self.options:
             if self.options[option]['required'] and option not in self.output_options.keys():
                 print(f'Error: option "{option}" is required')
@@ -399,4 +523,9 @@ class Arguments:
                         print(f'Error: operation "{required}" requires the option "{option}"')
                         sys.exit(1)
 
+        # Check if at least one operation is specified
+        if self.output_operations == {} and self.one_operation_required:
+            print('Error: one operation is required')
+            exit(1)
+
         return self.output_operations, self.output_options
